@@ -21,6 +21,7 @@ pub struct Backlight {
     current: Option<u64>,
     dbus: Option<Dbus>,
     has_write_permission: bool,
+    pending_dbus_write: bool,
 }
 
 impl Backlight {
@@ -37,6 +38,7 @@ impl Backlight {
                 .write(true)
                 .open(&brightness_path)?;
 
+            log::debug!("Using direct write on {} to change brightness value", path);
             (file, None)
         } else {
             let file = File::open(&brightness_path)?;
@@ -46,6 +48,18 @@ impl Backlight {
                 .and_then(|x| x.to_str())
                 .ok_or("Unable to identify backlight ID")?;
 
+            let subsystem = Path::new(path)
+                .parent()
+                .and_then(|p| p.file_name())
+                .and_then(|x| x.to_str())
+                .and_then(|x| match x {
+                    "backlight" | "leds" => Some(x),
+                    _ => None,
+                })
+                .ok_or(format!(
+                    "Unable to identify backlight subsystem out of {path}, please open an issue on GitHub"
+                ))?;
+
             let message = Message::new_method_call(
                 "org.freedesktop.login1",
                 "/org/freedesktop/login1/session/auto",
@@ -53,7 +67,7 @@ impl Backlight {
                 "SetBrightness",
             )
             .ok()
-            .map(|m| m.append2("backlight", id));
+            .map(|m| m.append2(subsystem, id));
 
             let connection = Connection::new_system().ok().and_then(|connection| {
                 message.map(|message| Dbus {
@@ -62,6 +76,7 @@ impl Backlight {
                 })
             });
 
+            log::debug!("Using DBUS for {} to change brightness value", path);
             (file, connection)
         };
 
@@ -87,6 +102,7 @@ impl Backlight {
             current: None,
             dbus,
             has_write_permission,
+            pending_dbus_write: false,
         })
     }
 }
@@ -103,10 +119,11 @@ impl super::Brightness for Backlight {
         match (self.inotify.read_events(&mut buffer), self.current) {
             (_, None) => update(self),
             (Ok(mut events), Some(cached)) => {
-                if events.next().is_some() {
-                    update(self)
-                } else {
+                if self.pending_dbus_write || events.next().is_none() {
+                    self.pending_dbus_write = false;
                     Ok(cached)
+                } else {
+                    update(self)
                 }
             }
             (Err(err), Some(cached)) if err.kind() == ErrorKind::WouldBlock => Ok(cached),
@@ -123,6 +140,7 @@ impl super::Brightness for Backlight {
             dbus.connection
                 .send(dbus.message.duplicate()?.append1(value as u32))
                 .map_err(|_| "Unable to send brightness change message via dbus")?;
+            self.pending_dbus_write = true;
         } else {
             Err(std::io::Error::from(ErrorKind::PermissionDenied))?
         }
