@@ -51,13 +51,15 @@ impl Controller {
     fn step(&mut self) {
         match self.brightness.get() {
             Ok(new_brightness) => {
+                let predicted_value = self.prediction_rx.try_iter().last();
+
                 // 1. check if user wants to learn a new value - this overrides any ongoing activity
                 if Some(new_brightness) != self.current {
                     return self.update_current(new_brightness);
                 }
 
                 // 2. check if predictor wants to set a new value
-                if let Some(desired) = self.prediction_rx.try_iter().last() {
+                if let Some(desired) = predicted_value {
                     self.update_target(desired);
                 }
 
@@ -87,9 +89,9 @@ impl Controller {
             (_, Some(current)) if desired == current => (),
             (_, Some(current)) => {
                 let step = if desired > current {
-                    div_ceil(desired - current, TRANSITION_MAX_MS)
+                    (desired - current).div_ceil(TRANSITION_MAX_MS) as i64
                 } else {
-                    -div_ceil(current - desired, TRANSITION_MAX_MS)
+                    -((current - desired).div_ceil(TRANSITION_MAX_MS) as i64)
                 };
                 self.target = Some(Target { desired, step });
             }
@@ -120,10 +122,6 @@ impl Controller {
     }
 }
 
-fn div_ceil(x: u64, y: u64) -> i64 {
-    ((x + y - 1) / y) as i64
-}
-
 #[cfg(test)]
 mod tests {
     use super::*;
