@@ -1,4 +1,3 @@
-mod controller;
+pub mod controller;
 mod data;
-
 pub use controller::Controller;
